from Device import *


class Control:
    def __init__(self, val=0):
        self.from_int(val)

    def from_int(self, val):
        self.count_div_sel = val & 0x03
        self.word_sel = (val >> 2) & 0x07
        self.tx_ctrl = (val >> 5) & 0x03
        self.rx_int_en = bool(val & 0x80)

    def to_int(self):
        val = self.count_div_sel
        val |= (self.word_sel << 2)
        val |= (self.tx_ctrl << 5)
        if self.rx_int_en:
            val |= 0x80
        return val

    def __repr__(self):
        ret = "CDS={self.count_div_sel}  WS={self.word_sel}  TC={self.tx_ctrl}  RIE={self.rx_int_en}".format(self=self)
        return ret


class Status:
    def __init__(self, val=0):
        self.from_int(val)

    def from_int(self, val):
        self.rx_data_full = bool(val & 1)
        self.tx_data_empty = bool(val & 2)
        self.dcd = bool(val & 4)
        self.cts = bool(val & 8)
        self.frame_err = bool(val & 0x10)
        self.rx_over = bool(val & 0x20)
        self.par_err = bool(val & 0x40)
        self.irq = bool(val & 0x80)

    def to_int(self):
        val = 0
        if self.rx_data_full:
            val |= 1
        if self.tx_data_empty:
            val |= 2
        if self.dcd:
            val |= 4
        if self.cts:
            val |= 8
        if self.frame_err:
            val |= 0x10
        if self.rx_over:
            val |= 0x20
        if self.par_err:
            val |= 0x40
        if self.irq:
            val |= 0x80
        return val

    def __repr__(self):
        ret = "RDRF={self.rx_data_full}  TDRE={self.tx_data_empty}  NDCD={self.dcd}  NCTS={self.cts}  FE={self.frame_err}  OVR={self.rx_over}  PE={self.par_err}  IRQ={self.irq}".format(self=self)
        return ret


class ACIA(Device):
    def __init__(self, system, addr_range, addr_bus_width):
        Device.__init__(self, addr_range, addr_bus_width)
        self.system = system
        self.data_in = []

    def reset(self):
        self.control = Control()
        self.status = Status(0x02)
        self.input = ""
        self.output = ""
        self.fileh = None

    def addInput(self, data_in, time=0):
        self.data_in.append((time, data_in))

    def readFrom(self, fileh):
        self.fileh = fileh

    def read(self, addr):
        reg = addr & 1
        if reg == 0:
            if self.trace:
                print("ACIA:", self.status)
            return self.status.to_int()

        if self.input:
            ret = self.input[0]
            self.input = self.input[1:]
            if self.trace:
                print("ACIA Rx:", repr(ret), repr(self.input))
            self.status.rx_data_full = (len(self.input) > 0)
            return ord(ret)
        print("** no data")
        return 0

    def write(self, addr, val):
        reg = addr & 1
        if reg == 0:
            self.control.from_int(val)
        else:
            self.output += chr(val)

    def tick(self, value):
        if self.input == "":
            if self.fileh:
                line = self.fileh.readline().rstrip('\r\n')
                if line:
                    self.input = line + '\r'
            elif self.data_in and value > self.data_in[0][0]:
                _, self.input = self.data_in.pop(0)
            self.status.rx_data_full = (len(self.input) > 0)

    def __repr__(self):
        ret = "Control:{self.control}  Status:{self.status}".format(self=self)
        ret += "\nOutput:" + repr(self.output)
        return ret
