from enum import Enum


class Break(Exception):
    pass


class CPU:
    def __init__(self, system):
        self.system = system
        self.instructions = {}
        self.breakpoints = {}
        self.trace = False

    def traceback(self):
        for addr, irq in self.return_stack:
            print("{:04x} {}".format(addr, 'IRQ' if irq else ''))

    def addBreak(self, addr, match=None):
        self.breakpoints[addr] = match

    def clearBreak(self, addr):
        self.breakpoints.pop(addr, None)

    def isBreakpoint(self, addr):
        if addr not in self.breakpoints:
            return False

        match = self.breakpoints[addr]
        if match is None:
            self.system.halt = True
            return True

        if callable(match):
            if match(self.system):
                self.system.halt = True
                return True
        else:
            if 'trace' in match:
                self.trace = match['trace']
            if 'halt' in match:
                self.system.halt = match['halt']
                return True
        return False

    def addrInDevice(self, addr):
        return False


class Instruction:
    def __init__(self, func, mode):
        self.func = func
        self.mode = mode

    def mnemonic(self):
        return self.func.__name__.rstrip('_')

    def __repr__(self):
        return "{} {}".format(self.mnemonic(), self.mode)


class Status:
    def __init__(self, val=0):
        self.from_int(val)

    def from_int(self, val):
        self.carry = bool(val & 1)
        self.overflow = bool(val & 2)
        self.zero = bool(val & 4)
        self.negative = bool(val & 8)
        self.int_mask = bool(val & 0x10)
        self.half_carry = bool(val & 0x20)

    def to_int(self):
        val = 0xC0
        if self.carry:
            val |= 1
        if self.overflow:
            val |= 2
        if self.zero:
            val |= 4
        if self.negative:
            val |= 8
        if self.int_mask:
            val |= 0x10
        if self.half_carry:
            val |= 0x20

        return val

    def __repr__(self):
        ret = "11"
        ret += 'H' if self.half_carry else '-'
        ret += 'I' if self.int_mask else '-'
        ret += 'N' if self.negative else '-'
        ret += 'Z' if self.zero else '-'
        ret += 'V' if self.overflow else '-'
        ret += 'C' if self.carry else '-'
        return ret


class Registers:
    def __init__(self):
        self.acc_a = 0
        self.acc_b = 0
        self.ind_x = 0
        self.stack_ptr = 0
        self.status = Status(0x10)
        self.ip = 0

    def __repr__(self):
        ret = "IP={self.ip:04x}  ACCA={self.acc_a:02x}  ACCB={self.acc_b:02x}  IX={self.ind_x:04x}  SP={self.stack_ptr:04x}  SR={self.status}".format(self=self)
        return ret


class M6802(CPU):

    class Mode(Enum):
        acc_a = 1
        acc_b = 2
        dir = 3
        ext = 4
        idx = 5
        imm = 6
        imm2 = 7
        inh = 8
        rel = 9

    def __init__(self, system):
        row0 = ((0x01, self.nop), (0x06, self.tap), (0x07, self.tpa), (0x08, self.inx),
                (0x09, self.dex), (0x0A, self.clv), (0x0B, self.sev), (0x0C, self.clc),
                (0x0D, self.sec), (0x0E, self.cli), (0x0F, self.sei))
        row1 = ((0x10, self.sba), (0x11, self.cba), (0x16, self.tab), (0x17, self.tba),
                (0x19, self.daa), (0x1B, self.aba))
        row2 = ((0x20, self.bra), (0x22, self.bhi), (0x23, self.bls), (0x24, self.bcc),
                (0x25, self.bcs), (0x26, self.bne), (0x27, self.beq), (0x28, self.bvc),
                (0x29, self.bvs), (0x2A, self.bpl), (0x2B, self.bmi), (0x2C, self.bge),
                (0x2D, self.blt), (0x2E, self.bgt), (0x2F, self.ble))
        row3 = ((0x30, self.tsx, self.Mode.inh), (0x31, self.ins, self.Mode.inh),
                (0x32, self.pul, self.Mode.acc_a), (0x33, self.pul, self.Mode.acc_b),
                (0x34, self.des, self.Mode.inh), (0x35, self.txs, self.Mode.inh),
                (0x36, self.psh, self.Mode.acc_a), (0x37, self.psh, self.Mode.acc_b),
                (0x39, self.rts, self.Mode.inh), (0x3B, self.rti, self.Mode.inh),
                (0x3E, self.wai, self.Mode.inh), (0x3F, self.swi, self.Mode.inh))
        row47 = ((0x40, self.Mode.acc_a), (0x50, self.Mode.acc_b),
                 (0x60, self.Mode.idx), (0x70, self.Mode.ext))
        col47 = ((0x00, self.neg), (0x03, self.com), (0x04, self.lsr), (0x06, self.ror),
                 (0x07, self.asr), (0x08, self.asl), (0x09, self.rol), (0x0A, self.dec),
                 (0x0C, self.inc), (0x0D, self.tst), (0x0F, self.clr))
        row8F = ((0x80, self.Mode.imm), (0x90, self.Mode.dir), (0xA0, self.Mode.idx), (0xB0, self.Mode.ext),
                 (0xC0, self.Mode.imm), (0xD0, self.Mode.dir), (0xE0, self.Mode.idx), (0xF0, self.Mode.ext))
        col8F = ((0x00, self.sub), (0x01, self.cmp_), (0x02, self.sbc), (0x04, self.and_),
                 (0x05, self.bit), (0x06, self.lda), (0x08, self.eor), (0x09, self.adc),
                 (0x0A, self.ora), (0x0B, self.add))
        rowSTA = ((0x97, self.Mode.dir), (0xA7, self.Mode.idx), (0xB7, self.Mode.ext),
                  (0xD7, self.Mode.dir), (0xE7, self.Mode.idx), (0xF7, self.Mode.ext))
        rowCPX = ((0x8C, self.Mode.imm2), (0x9C, self.Mode.dir), (0xAC, self.Mode.idx), (0xBC, self.Mode.ext))
        rowLDS = ((0x8E, self.Mode.imm2), (0x9E, self.Mode.dir), (0xAE, self.Mode.idx), (0xBE, self.Mode.ext))
        rowLDX = ((0xCE, self.Mode.imm2), (0xDE, self.Mode.dir), (0xEE, self.Mode.idx), (0xFE, self.Mode.ext))
        rowST = ((0x90, self.Mode.dir), (0xA0, self.Mode.idx), (0xB0, self.Mode.ext))

        CPU.__init__(self, system)

        for opcode, func in row0 + row1:
            self.instructions[opcode] = Instruction(func, self.Mode.inh)
        for opcode, func in row2:
            self.instructions[opcode] = Instruction(func, self.Mode.rel)
        for opcode, func, mode in row3:
            self.instructions[opcode] = Instruction(func, mode)
        for upper, mode in row47:
            for lower, func in col47:
                self.instructions[upper+lower] = Instruction(func, mode)
        self.instructions[0x6E] = Instruction(self.jmp, self.Mode.idx)
        self.instructions[0x7E] = Instruction(self.jmp, self.Mode.ext)
        for upper, mode in row8F:
            for lower, func in col8F:
                self.instructions[upper+lower] = Instruction(func, mode)
        for opcode, mode in rowSTA:
            self.instructions[opcode] = Instruction(self.sta, mode)
        for opcode, mode in rowCPX:
            self.instructions[opcode] = Instruction(self.cpx, mode)
        for opcode, mode in rowLDS:
            self.instructions[opcode] = Instruction(self.lds, mode)
        for opcode, mode in rowLDX:
            self.instructions[opcode] = Instruction(self.ldx, mode)
        for opcode, mode in rowST:
            self.instructions[opcode+0xF] = Instruction(self.sts, mode)
            self.instructions[opcode+0x4F] = Instruction(self.stx, mode)
        self.instructions[0x8D] = Instruction(self.bsr, self.Mode.rel)
        self.instructions[0xAD] = Instruction(self.jsr, self.Mode.idx)
        self.instructions[0xBD] = Instruction(self.jsr, self.Mode.ext)

    def reset(self):
        self.reg = Registers()
        self.reg.ip = self.system.read_be_word(0xFFFE)
        self.irq_pending = False
        self.return_stack = []

    def next(self):
        self.system.halt = False
        trace = self.trace
        start = len(self.return_stack)
        while not self.system.halt:
            self.step()
            self.trace = False
            if len(self.return_stack) <= start:
                break
        self.trace = trace

    def step(self, checkBreak=False):
        if self.trace:
            self.where()

        if self.irq_pending and not self.reg.status.int_mask:
            if self.trace:
                print("** IRQ")
            self.int_save_reg()
            self.irq_pending = False
            self.reg.status.int_mask = True
            self.reg.ip = self.system.read_be_word(0xFFF8)
            self.return_stack.append((self.reg.ip, True))
            return

        if checkBreak and self.isBreakpoint(self.reg.ip):
            print(self, "\n** Break")
            return

        self.system.tick()
        self.opcode = self.system.read(self.reg.ip)
        if self.opcode not in self.instructions:
            print(self)
            raise NotImplemented("{:02x} at {:04x}".format(self.opcode, self.reg.ip))
        self.reg.ip += 1
        inst = self.instructions[self.opcode]
        if inst.mode in (self.Mode.ext, self.Mode.imm2):
            val = self.system.read_be_word(self.reg.ip)
            self.reg.ip += 2
            inst.func(inst.mode, val)
        elif inst.mode in (self.Mode.dir, self.Mode.idx, self.Mode.imm):
            val = self.system.read(self.reg.ip)
            self.reg.ip += 1
            inst.func(inst.mode, val)
        elif inst.mode == self.Mode.rel:
            val = self.system.read(self.reg.ip)
            self.reg.ip += 1
            inst.func(val)
        elif inst.mode in (self.Mode.acc_a, self.Mode.acc_b):
            inst.func(inst.mode)
        else:
            inst.func()
        self.system.tick()

    def where(self):
        opcode = self.system.read(self.reg.ip)
        if opcode not in self.instructions:
            print("??? {:02x} at {:04x}".format(opcode, self.reg.ip))
            return
        inst = self.instructions[opcode]
        text = inst.mnemonic()
        if opcode >= 0x80 and (opcode & 0xF) <= 0xC:
            if opcode >= 0xC0:
                text += ' b'
            else:
                text += ' a'
        if inst.mode == self.Mode.ext:
            val = self.system.read_be_word(self.reg.ip+1)
            text += " ${:04x}".format(val)
        elif inst.mode == self.Mode.imm2:
            val = self.system.read_be_word(self.reg.ip+1)
            text += " #${:04x}".format(val)
        elif inst.mode == self.Mode.dir:
            val = self.system.read(self.reg.ip+1)
            text += " ${:02x}".format(val)
        elif inst.mode == self.Mode.imm:
            val = self.system.read(self.reg.ip+1)
            text += " #${:02x}".format(val)
        elif inst.mode == self.Mode.idx:
            val = self.system.read(self.reg.ip+1)
            text += " ${:02x},x".format(val)
        elif inst.mode == self.Mode.rel:
            val = self.system.read(self.reg.ip+1)
            text += " ${:04x}".format(self.reg.ip+2+self.ip_offset(val))
        elif inst.mode == self.Mode.acc_a:
            text += " a"
        elif inst.mode == self.Mode.acc_b:
            text += " b"
        print("{:04x}  {:22} {}".format(self.reg.ip, text, self))

    def __repr__(self):
        return self.reg.__repr__()

    def base_add(self, val1, val2, carry=False):
        res = (val1+val2+int(carry)) & 0xFF
        self.reg.status.negative = bool(res & 0x80)
        self.reg.status.zero = (res == 0)
        self.reg.status.overflow = bool(((val1 & val2 & ~res) | (~val1 & ~val2 & res)) & 0x80)
        self.reg.status.carry = bool(((val1 & val2) | (val2 & ~res) | (~val2 & res)) & 0x80)
        self.reg.status.half_carry = bool(((val1 & val2) | (val2 & ~res) | (~res & val1)) & 0x10)
        return res

    def subtract(self, val1, val2, carry=False):
        res = (val1-val2-int(carry)) & 0xFF

        self.reg.status.negative = bool(res & 0x80)
        self.reg.status.zero = (res == 0)
        self.reg.status.overflow = bool(((val1 & ~val2 & ~res) | (~val1 & val2 & res)) & 0x80)
        self.reg.status.carry = bool(((~val1 & val2) | (val2 & res) | (res & ~val1)) & 0x80)
        return res

    def update_szv(self, val):
        self.reg.status.negative = bool(val & 0x80)
        self.reg.status.zero = (val == 0)
        self.reg.status.overflow = False

    def update_szv_word(self, val):
        self.reg.status.negative = bool(val & 0x8000)
        self.reg.status.zero = (val == 0)
        self.reg.status.overflow = False

    def ip_offset(self, val):
        if val >= 0x80:
            return -(0x100-val)
        return val

    def int_save_reg(self):
        self.system.tick()
        self.system.write_be_word(self.reg.stack_ptr-1, self.reg.ip)
        self.system.tick()
        self.system.write_be_word(self.reg.stack_ptr-3, self.reg.ind_x)
        self.system.tick()
        self.system.write(self.reg.stack_ptr-4, self.reg.acc_a)
        self.system.tick()
        self.system.write(self.reg.stack_ptr-5, self.reg.acc_b)
        self.system.tick()
        self.system.write(self.reg.stack_ptr-6, self.reg.status.to_int())
        self.system.tick()
        self.reg.stack_ptr = (self.reg.stack_ptr-7) & 0xFFFF
        self.system.tick()

    def nop(self):
        pass

    def tap(self):
        self.reg.status.from_int(self.reg.acc_a)

    def tpa(self):
        self.reg.acc_a = self.reg.status.to_int()

    def inx(self):
        self.reg.ind_x = (self.reg.ind_x+1) & 0xFFFF
        self.system.tick()
        self.reg.status.zero = (self.reg.ind_x == 0)
        self.system.tick()

    def dex(self):
        self.reg.ind_x = (self.reg.ind_x-1) & 0xFFFF
        self.system.tick()
        self.reg.status.zero = (self.reg.ind_x == 0)
        self.system.tick()

    def clv(self):
        self.reg.status.overflow = False

    def sev(self):
        self.reg.status.overflow = True

    def clc(self):
        self.reg.status.carry = False

    def sec(self):
        self.reg.status.carry = True

    def cli(self):
        self.reg.status.int_mask = False

    def sei(self):
        self.reg.status.int_mask = True

    def sba(self):
        self.reg.acc_a = self.subtract(self.reg.acc_a, self.reg.acc_b)

    def cba(self):
        self.subtract(self.reg.acc_a, self.reg.acc_b)

    def tab(self):
        self.reg.acc_b = self.reg.acc_a
        self.update_szv(self.reg.acc_b)

    def tba(self):
        self.reg.acc_a = self.reg.acc_b
        self.update_szv(self.reg.acc_a)

    def daa(self):
        if self.reg.status.half_carry or (self.reg.acc_a & 0xF) >= 0xA:
            self.reg.acc_a += 0x6
        if self.reg.status.carry or (self.reg.acc_a & 0xF0) >= 0xA0:
            self.reg.acc_a += 0x60
            self.reg.status.carry = True
        self.reg.acc_a = self.reg.acc_a & 0xFF
        self.update_szv(self.reg.acc_a)

    def aba(self):
        self.reg.acc_a = self.base_add(self.reg.acc_a, self.reg.acc_b)

    def bra(self, off):
        self.system.tick()
        self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def bhi(self, off):
        self.system.tick()
        if not (self.reg.status.carry or self.reg.status.zero):
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def bls(self, off):
        self.system.tick()
        if self.reg.status.carry or self.reg.status.zero:
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def bcc(self, off):
        self.system.tick()
        if not self.reg.status.carry:
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def bcs(self, off):
        self.system.tick()
        if self.reg.status.carry:
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def bne(self, off):
        self.system.tick()
        if not self.reg.status.zero:
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def beq(self, off):
        self.system.tick()
        if self.reg.status.zero:
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def bvc(self, off):
        self.system.tick()
        if not self.reg.status.overflow:
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def bvs(self, off):
        self.system.tick()
        if self.reg.status.overflow:
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def bpl(self, off):
        self.system.tick()
        if not self.reg.status.negative:
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def bmi(self, off):
        self.system.tick()
        if self.reg.status.negative:
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def bge(self, off):
        self.system.tick()
        if not(self.reg.status.negative ^ self.reg.status.overflow):
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def blt(self, off):
        self.system.tick()
        if self.reg.status.negative ^ self.reg.status.overflow:
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def bgt(self, off):
        self.system.tick()
        if not(self.reg.status.zero | (self.reg.status.negative ^ self.reg.status.overflow)):
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def ble(self, off):
        self.system.tick()
        if self.reg.status.zero | (self.reg.status.negative ^ self.reg.status.overflow):
            self.reg.ip += self.ip_offset(off)
        self.system.tick()

    def tsx(self):
        self.system.tick()
        self.reg.ind_x = (self.reg.stack_ptr+1) & 0xFFFF
        self.system.tick()

    def ins(self):
        self.system.tick()
        self.reg.stack_ptr = (self.reg.stack_ptr+1) & 0xFFFF
        self.system.tick()

    def pul(self, mode):
        self.system.tick()
        self.reg.stack_ptr = (self.reg.stack_ptr+1) & 0xFFFF
        if mode == self.Mode.acc_a:
            self.reg.acc_a = self.system.read(self.reg.stack_ptr)
        else:
            self.reg.acc_b = self.system.read(self.reg.stack_ptr)
        self.system.tick()

    def des(self):
        self.system.tick()
        self.reg.stack_ptr = (self.reg.stack_ptr-1) & 0xFFFF
        self.system.tick()

    def txs(self):
        self.system.tick()
        self.reg.stack_ptr = (self.reg.ind_x-1) & 0xFFFF
        self.system.tick()

    def psh(self, mode):
        self.system.tick()
        if mode == self.Mode.acc_a:
            self.system.write(self.reg.stack_ptr, self.reg.acc_a)
        else:
            self.system.write(self.reg.stack_ptr, self.reg.acc_b)
        self.reg.stack_ptr = (self.reg.stack_ptr-1) & 0xFFFF
        self.system.tick()

    def rts(self):
        self.system.tick()
        self.reg.ip = self.system.read_be_word(self.reg.stack_ptr+1)
        self.system.tick()
        self.reg.stack_ptr = (self.reg.stack_ptr+2) & 0xFFFF
        self.system.tick()
        self.return_stack.pop()

    def rti(self):
        self.system.tick()
        self.reg.status.from_int(self.system.read(self.reg.stack_ptr+1))
        self.system.tick()
        self.reg.acc_b = self.system.read(self.reg.stack_ptr+2)
        self.system.tick()
        self.reg.acc_a = self.system.read(self.reg.stack_ptr+3)
        self.system.tick()
        self.reg.ind_x = self.system.read_be_word(self.reg.stack_ptr+4)
        self.system.tick()
        self.reg.ip = self.system.read_be_word(self.reg.stack_ptr+6)
        self.system.tick()
        self.reg.stack_ptr = (self.reg.stack_ptr+7) & 0xFFFF
        self.system.tick()
        self.return_stack.pop()
        self.system.tick()

    def wai(self):
        self.int_save_reg()
        raise NotImplementedError("Wait for INT")

    def swi(self):
        self.int_save_reg()
        self.system.tick()
        self.reg.ip = self.system.read_be_word(0xFFFA)
        self.system.tick()
        self.system.tick()

    def neg(self, mode, val=0):
        if mode == self.Mode.acc_a:
            init = self.reg.acc_a
            res = 0x100-init
            self.reg.acc_a = res
        elif mode == self.Mode.acc_b:
            init = self.reg.acc_b
            res = 0x100-init
            self.reg.acc_b = res
        elif mode == self.Mode.idx:
            self.system.tick()
            self.system.tick()
            init = self.system.read(val+self.reg.ind_x)
            self.system.tick()
            res = 0x100-init
            self.system.tick()
            self.system.write(val+self.reg.ind_x, res)
            self.system.tick()
        elif mode == self.Mode.ext:
            self.system.tick()
            init = self.system.read(val)
            self.system.tick()
            res = 0x100-init
            self.system.tick()
            self.system.write(val, res)
            self.system.tick()

        self.reg.status.overflow = bool((init & res) & 0x80)
        self.reg.status.negative = bool(res & 0x80)
        self.reg.status.zero = (res == 0)
        self.reg.status.carry = self.reg.status.zero

    def com(self, mode, val=0):
        if mode == self.Mode.acc_a:
            init = self.reg.acc_a
            res = 0xFF-init
            self.reg.acc_a = res
        elif mode == self.Mode.acc_b:
            init = self.reg.acc_b
            res = 0xFF-init
            self.reg.acc_b = res
        elif mode == self.Mode.idx:
            self.system.tick()
            self.system.tick()
            init = self.system.read(val+self.reg.ind_x)
            self.system.tick()
            res = 0xFF-init
            self.system.tick()
            self.system.write(val+self.reg.ind_x, res)
            self.system.tick()
        elif mode == self.Mode.ext:
            self.system.tick()
            init = self.system.read(val)
            self.system.tick()
            res = 0xFF-init
            self.system.tick()
            self.system.write(val, res)
            self.system.tick()

        self.reg.status.overflow = False
        self.reg.status.negative = bool(res & 0x80)
        self.reg.status.zero = (res == 0)
        self.reg.status.carry = True

    def lsr(self, mode, val=0):
        if mode == self.Mode.acc_a:
            init = self.reg.acc_a
            res = init >> 1
            self.reg.acc_a = res
        elif mode == self.Mode.acc_b:
            init = self.reg.acc_b
            res = init >> 1
            self.reg.acc_b = res
        elif mode == self.Mode.idx:
            self.system.tick()
            self.system.tick()
            init = self.system.read(val+self.reg.ind_x)
            self.system.tick()
            res = init >> 1
            self.system.tick()
            self.system.write(val+self.reg.ind_x, res)
            self.system.tick()
        elif mode == self.Mode.ext:
            self.system.tick()
            init = self.system.read(val)
            self.system.tick()
            res = init >> 1
            self.system.tick()
            self.system.write(val, res)
            self.system.tick()

        self.reg.status.negative = bool(res & 0x80)
        self.reg.status.zero = (res == 0)
        self.reg.status.carry = bool(init & 1)
        self.reg.status.overflow = self.reg.status.carry

    def ror(self, mode, val=0):
        if mode == self.Mode.acc_a:
            init = self.reg.acc_a
            res = (init >> 1) | 0x80 if self.reg.status.carry else 0
            self.reg.acc_a = res
        elif mode == self.Mode.acc_b:
            init = self.reg.acc_b
            res = (init >> 1) | 0x80 if self.reg.status.carry else 0
            self.reg.acc_b = res
        elif mode == self.Mode.idx:
            self.system.tick()
            self.system.tick()
            init = self.system.read(val+self.reg.ind_x)
            self.system.tick()
            res = (init >> 1) | 0x80 if self.reg.status.carry else 0
            self.system.tick()
            self.system.write(val+self.reg.ind_x, res)
            self.system.tick()
        elif mode == self.Mode.ext:
            self.system.tick()
            init = self.system.read(val)
            self.system.tick()
            res = (init >> 1) | 0x80 if self.reg.status.carry else 0
            self.system.tick()
            self.system.write(val, res)
            self.system.tick()

        self.reg.status.negative = bool(res & 0x80)
        self.reg.status.zero = (res == 0)
        self.reg.status.carry = bool(init & 1)
        self.reg.status.overflow = self.reg.status.negative ^ self.reg.status.carry

    def asr(self, mode, val):
        if mode == self.Mode.acc_a:
            init = self.reg.acc_a
            res = (init >> 1) | (init & 0x80)
            self.reg.acc_a = res
        elif mode == self.Mode.acc_b:
            init = self.reg.acc_b
            res = (init >> 1) | (init & 0x80)
            self.reg.acc_b = res
        elif mode == self.Mode.idx:
            self.system.tick()
            self.system.tick()
            init = self.system.read(val+self.reg.ind_x)
            self.system.tick()
            res = (init >> 1) | (init & 0x80)
            self.system.tick()
            self.system.write(val+self.reg.ind_x, res)
            self.system.tick()
        elif mode == self.Mode.ext:
            self.system.tick()
            init = self.system.read(val)
            self.system.tick()
            res = (init >> 1) | (init & 0x80)
            self.system.tick()
            self.system.write(val, res)
            self.system.tick()

        self.reg.status.negative = bool(res & 0x80)
        self.reg.status.zero = (res == 0)
        self.reg.status.carry = bool(init & 1)
        self.reg.status.overflow = self.reg.status.negative ^ self.reg.status.carry

    def asl(self, mode, val=0):
        if mode == self.Mode.acc_a:
            init = self.reg.acc_a
            res = (init << 1) & 0xFF
            self.reg.acc_a = res
        elif mode == self.Mode.acc_b:
            init = self.reg.acc_b
            res = (init << 1) & 0xFF
            self.reg.acc_b = res
        elif mode == self.Mode.idx:
            self.system.tick()
            self.system.tick()
            init = self.system.read(val+self.reg.ind_x)
            self.system.tick()
            res = (init << 1) & 0xFF
            self.system.tick()
            self.system.write(val+self.reg.ind_x, res)
            self.system.tick()
        elif mode == self.Mode.ext:
            self.system.tick()
            init = self.system.read(val)
            self.system.tick()
            res = (init << 1) & 0xFF
            self.system.tick()
            self.system.write(val, res)
            self.system.tick()

        self.reg.status.negative = bool(res & 0x80)
        self.reg.status.zero = (res == 0)
        self.reg.status.carry = bool(init & 0x80)
        self.reg.status.overflow = self.reg.status.negative ^ self.reg.status.carry

    def rol(self, mode, val=0):
        if mode == self.Mode.acc_a:
            init = self.reg.acc_a
            res = (init << 1) & 0xFF | 1 if self.reg.status.carry else 0
            self.reg.acc_a = res
        elif mode == self.Mode.acc_b:
            init = self.reg.acc_b
            res = (init << 1) & 0xFF | 1 if self.reg.status.carry else 0
            self.reg.acc_b = res
        elif mode == self.Mode.idx:
            self.system.tick()
            self.system.tick()
            init = self.system.read(val+self.reg.ind_x)
            self.system.tick()
            res = (init << 1) & 0xFF | 1 if self.reg.status.carry else 0
            self.system.tick()
            self.system.write(val+self.reg.ind_x, res)
            self.system.tick()
        elif mode == self.Mode.ext:
            self.system.tick()
            init = self.system.read(val)
            self.system.tick()
            res = (init << 1) & 0xFF | 1 if self.reg.status.carry else 0
            self.system.tick()
            self.system.write(val, res)
            self.system.tick()

        self.reg.status.negative = bool(res & 0x80)
        self.reg.status.zero = (res == 0)
        self.reg.status.carry = bool(init & 0x80)
        self.reg.status.overflow = self.reg.status.negative ^ self.reg.status.carry

    def dec(self, mode, val=0):
        if mode == self.Mode.acc_a:
            init = self.reg.acc_a
            res = (init-1) & 0xFF
            self.reg.acc_a = res
        elif mode == self.Mode.acc_b:
            init = self.reg.acc_b
            res = (init-1) & 0xFF
            self.reg.acc_b = res
        elif mode == self.Mode.idx:
            self.system.tick()
            self.system.tick()
            init = self.system.read(val+self.reg.ind_x)
            self.system.tick()
            res = (init-1) & 0xFF
            self.system.tick()
            self.system.write(val+self.reg.ind_x, res)
            self.system.tick()
        elif mode == self.Mode.ext:
            self.system.tick()
            init = self.system.read(val)
            self.system.tick()
            res = (init-1) & 0xFF
            self.system.tick()
            self.system.write(val, res)
            self.system.tick()

        self.reg.status.negative = bool(res & 0x80)
        self.reg.status.zero = (res == 0)
        self.reg.status.overflow = ~res & init

    def inc(self, mode, val=0):
        if mode == self.Mode.acc_a:
            init = self.reg.acc_a
            res = (init+1) & 0xFF
            self.reg.acc_a = res
        elif mode == self.Mode.acc_b:
            init = self.reg.acc_b
            res = (init+1) & 0xFF
            self.reg.acc_b = res
        elif mode == self.Mode.idx:
            self.system.tick()
            self.system.tick()
            init = self.system.read(val+self.reg.ind_x)
            self.system.tick()
            res = (init+1) & 0xFF
            self.system.tick()
            self.system.write(val+self.reg.ind_x, res)
            self.system.tick()
        elif mode == self.Mode.ext:
            self.system.tick()
            init = self.system.read(val)
            self.system.tick()
            res = (init+1) & 0xFF
            self.system.tick()
            self.system.write(val, res)
            self.system.tick()

        self.reg.status.negative = bool(res & 0x80)
        self.reg.status.zero = (res == 0)
        self.reg.status.overflow = res & ~init

    def tst(self, mode, val=0):
        if mode == self.Mode.acc_a:
            init = self.reg.acc_a
        elif mode == self.Mode.acc_b:
            init = self.reg.acc_b
        elif mode == self.Mode.idx:
            self.system.tick()
            self.system.tick()
            init = self.system.read(val+self.reg.ind_x)
            self.system.tick()
            self.system.tick()
            self.system.tick()
        elif mode == self.Mode.ext:
            self.system.tick()
            self.system.tick()
            init = self.system.read(val)
            self.system.tick()
            self.system.tick()

        self.reg.status.negative = bool(init & 0x80)
        self.reg.status.zero = (init == 0)
        self.reg.status.overflow = False
        self.reg.status.carry = False

    def clr(self, mode, val=0):
        if mode == self.Mode.acc_a:
            self.reg.acc_a = 0
        elif mode == self.Mode.acc_b:
            self.reg.acc_b = 0
        elif mode == self.Mode.idx:
            self.system.tick()
            self.system.tick()
            self.system.write(val+self.reg.ind_x, 0)
            self.system.tick()
            self.system.tick()
            self.system.tick()
        elif mode == self.Mode.ext:
            self.system.tick()
            self.system.tick()
            self.system.write(val, 0)
            self.system.tick()
            self.system.tick()

        self.reg.status.negative = False
        self.reg.status.zero = True
        self.reg.status.overflow = False
        self.reg.status.carry = False

    def jmp(self, mode, val):
        if mode == self.Mode.idx:
            self.reg.ip = self.system.read(val+self.reg.ind_x)
            self.system.tick()
        elif mode == self.Mode.ext:
            self.reg.ip = val
        self.system.tick()

    def sub(self, mode, val):
        acc_b = bool(self.opcode & 0x40)
        if mode == self.Mode.imm:
            if acc_b:
                self.reg.acc_b = self.subtract(self.reg.acc_b, val)
            else:
                self.reg.acc_a = self.subtract(self.reg.acc_a, val)
        elif mode in (self.Mode.dir, self.Mode.ext):
            if acc_b:
                self.reg.acc_b = self.subtract(self.reg.acc_b, self.system.read(val))
            else:
                self.reg.acc_a = self.subtract(self.reg.acc_a, self.system.read(val))
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            if acc_b:
                self.reg.acc_b = self.subtract(self.reg.acc_b, self.system.read(val+self.reg.ind_x))
            else:
                self.reg.acc_a = self.subtract(self.reg.acc_a, self.system.read(val+self.reg.ind_x))

    def cmp_(self, mode, val):
        acc_b = bool(self.opcode & 0x40)
        if mode == self.Mode.imm:
            if acc_b:
                self.subtract(self.reg.acc_b, val)
            else:
                self.subtract(self.reg.acc_a, val)
        elif mode in (self.Mode.dir, self.Mode.ext):
            if acc_b:
                self.subtract(self.reg.acc_b, self.system.read(val))
            else:
                self.subtract(self.reg.acc_a, self.system.read(val))
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            if acc_b:
                self.subtract(self.reg.acc_b, self.system.read(val+self.reg.ind_x))
            else:
                self.subtract(self.reg.acc_a, self.system.read(val+self.reg.ind_x))

    def sbc(self, mode, val):
        acc_b = bool(self.opcode & 0x40)
        if mode == self.Mode.imm:
            if acc_b:
                self.reg.acc_b = self.subtract(self.reg.acc_b, val, self.reg.status.carry)
            else:
                self.reg.acc_a = self.subtract(self.reg.acc_a, val, self.reg.status.carry)
        elif mode in (self.Mode.dir, self.Mode.ext):
            if acc_b:
                self.reg.acc_b = self.subtract(self.reg.acc_b, self.system.read(val), self.reg.status.carry)
            else:
                self.reg.acc_a = self.subtract(self.reg.acc_a, self.system.read(val), self.reg.status.carry)
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            if acc_b:
                self.reg.acc_b = self.subtract(self.reg.acc_b, self.system.read(val+self.reg.ind_x), self.reg.status.carry)
            else:
                self.reg.acc_a = self.subtract(self.reg.acc_a, self.system.read(val+self.reg.ind_x), self.reg.status.carry)

    def and_(self, mode, val):
        acc_b = bool(self.opcode & 0x40)
        if mode == self.Mode.imm:
            if acc_b:
                res = self.reg.acc_b & val
                self.reg.acc_b = res
            else:
                res = self.reg.acc_a & val
                self.reg.acc_a = res
        elif mode in (self.Mode.dir, self.Mode.ext):
            if acc_b:
                res = self.reg.acc_b & self.system.read(val)
                self.reg.acc_b = res
            else:
                res = self.reg.acc_a & self.system.read(val)
                self.reg.acc_a = res
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            if acc_b:
                res = self.reg.acc_b & self.system.read(val+self.reg.ind_x)
                self.reg.acc_b = res
            else:
                res = self.reg.acc_a & self.system.read(val+self.reg.ind_x)
                self.reg.acc_a = res
        self.update_szv(res)

    def bit(self, mode, val):
        acc_b = bool(self.opcode & 0x40)
        if mode == self.Mode.imm:
            if acc_b:
                res = self.reg.acc_b & val
            else:
                res = self.reg.acc_a & val
        elif mode in (self.Mode.dir, self.Mode.ext):
            if acc_b:
                res = self.reg.acc_b & self.system.read(val)
            else:
                res = self.reg.acc_a & self.system.read(val)
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            if acc_b:
                res = self.reg.acc_b & self.system.read(val+self.reg.ind_x)
            else:
                res = self.reg.acc_a & self.system.read(val+self.reg.ind_x)
        self.update_szv(res)

    def lda(self, mode, val):
        acc_b = bool(self.opcode & 0x40)
        if mode == self.Mode.imm:
            res = val
        elif mode in (self.Mode.dir, self.Mode.ext):
            res = self.system.read(val)
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            res = self.system.read(val+self.reg.ind_x)
        if acc_b:
            self.reg.acc_b = res
        else:
            self.reg.acc_a = res
        self.update_szv(res)

    def eor(self, mode, val):
        acc_b = bool(self.opcode & 0x40)
        if mode == self.Mode.imm:
            if acc_b:
                res = self.reg.acc_b ^ val
                self.reg.acc_b = res
            else:
                res = self.reg.acc_a ^ val
                self.reg.acc_a = res
        elif mode in (self.Mode.dir, self.Mode.ext):
            if acc_b:
                res = self.reg.acc_b ^ self.system.read(val)
                self.reg.acc_b = res
            else:
                res = self.reg.acc_a ^ self.system.read(val)
                self.reg.acc_a = res
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            if acc_b:
                res = self.reg.acc_b ^ self.system.read(val+self.reg.ind_x)
                self.reg.acc_b = res
            else:
                res = self.reg.acc_a ^ self.system.read(val+self.reg.ind_x)
                self.reg.acc_a = res
        self.update_szv(res)

    def adc(self, mode, val):
        acc_b = bool(self.opcode & 0x40)
        if mode == self.Mode.imm:
            if acc_b:
                self.reg.acc_b = self.base_add(self.reg.acc_b, val, self.reg.status.carry)
            else:
                self.reg.acc_a = self.base_add(self.reg.acc_a, val, self.reg.status.carry)
        elif mode in (self.Mode.dir, self.Mode.ext):
            if acc_b:
                self.reg.acc_b = self.base_add(self.reg.acc_b, self.system.read(val), self.reg.status.carry)
            else:
                self.reg.acc_a = self.base_add(self.reg.acc_a, self.system.read(val), self.reg.status.carry)
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            if acc_b:
                self.reg.acc_b = self.base_add(self.reg.acc_b, self.system.read(val+self.reg.ind_x), self.reg.status.carry)
            else:
                self.reg.acc_a = self.base_add(self.reg.acc_a, self.system.read(val+self.reg.ind_x), self.reg.status.carry)

    def ora(self, mode, val):
        acc_b = bool(self.opcode & 0x40)
        if mode == self.Mode.imm:
            if acc_b:
                res = self.reg.acc_b | val
                self.reg.acc_b = res
            else:
                res = self.reg.acc_a | val
                self.reg.acc_a = res
        elif mode in (self.Mode.dir, self.Mode.ext):
            if acc_b:
                res = self.reg.acc_b | self.system.read(val)
                self.reg.acc_b = res
            else:
                res = self.reg.acc_a | self.system.read(val)
                self.reg.acc_a = res
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            if acc_b:
                res = self.reg.acc_b | self.system.read(val+self.reg.ind_x)
                self.reg.acc_b = res
            else:
                res = self.reg.acc_a | self.system.read(val+self.reg.ind_x)
                self.reg.acc_a = res
        self.update_szv(res)

    def add(self, mode, val):
        acc_b = bool(self.opcode & 0x40)
        if mode == self.Mode.imm:
            if acc_b:
                self.reg.acc_b = self.base_add(self.reg.acc_b, val)
            else:
                self.reg.acc_a = self.base_add(self.reg.acc_a, val)
        elif mode in (self.Mode.dir, self.Mode.ext):
            if acc_b:
                self.reg.acc_b = self.base_add(self.reg.acc_b, self.system.read(val))
            else:
                self.reg.acc_a = self.base_add(self.reg.acc_a, self.system.read(val))
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            if acc_b:
                self.reg.acc_b = self.base_add(self.reg.acc_b, self.system.read(val+self.reg.ind_x))
            else:
                self.reg.acc_a = self.base_add(self.reg.acc_a, self.system.read(val+self.reg.ind_x))

    def sta(self, mode, val):
        self.system.tick()
        acc_b = bool(self.opcode & 0x40)
        if acc_b:
            acc = self.reg.acc_b
        else:
            acc = self.reg.acc_a
        if mode in (self.Mode.dir, self.Mode.ext):
            self.system.tick()
            self.system.write(val, acc)
            self.system.tick()
        elif mode == self.Mode.idx:
            self.system.tick()
            self.system.write(val+self.reg.ind_x, acc)
            self.system.tick()
        self.update_szv(acc)
        self.system.tick()

    def cpx(self, mode, val):
        indx = self.reg.ind_x
        if mode == self.Mode.imm2:
            res = self.reg.ind_x - val
        elif mode in (self.Mode.dir, self.Mode.ext):
            self.system.tick()
            res = self.reg.ind_x - self.system.read_be_word(val)
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            self.system.tick()
            res = self.reg.ind_x - self.system.read_be_word(val+self.reg.ind_x)
            self.system.tick()
            self.system.tick()
        self.reg.status.negative = bool(res & 0x8000)
        self.reg.status.zero = (res == 0)
        self.reg.status.overflow = bool((res ^ indx) & (res ^ self.reg.ind_x) & 0x8000)
        self.system.tick()

    def lds(self, mode, val):
        if mode == self.Mode.imm2:
            res = val
        elif mode in (self.Mode.dir, self.Mode.ext):
            self.system.tick()
            res = self.system.read_be_word(val)
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            self.system.tick()
            res = self.system.read_be_word(val+self.reg.ind_x)
            self.system.tick()
            self.system.tick()
        self.reg.stack_ptr = res
        self.update_szv_word(res)
        self.system.tick()

    def ldx(self, mode, val):
        if mode == self.Mode.imm2:
            res = val
        elif mode in (self.Mode.dir, self.Mode.ext):
            self.system.tick()
            res = self.system.read_be_word(val)
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            self.system.tick()
            res = self.system.read_be_word(val+self.reg.ind_x)
            self.system.tick()
            self.system.tick()
        self.reg.ind_x = res
        self.update_szv_word(res)
        self.system.tick()

    def sts(self, mode, val):
        self.system.tick()
        if mode in (self.Mode.dir, self.Mode.ext):
            self.system.write_be_word(val, self.reg.stack_ptr)
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            self.system.write_be_word(val+self.reg.ind_x, self.reg.stack_ptr)
            self.system.tick()
            self.system.tick()
        self.update_szv_word(self.reg.stack_ptr)
        self.system.tick()
        self.system.tick()

    def stx(self, mode, val):
        self.system.tick()
        if mode in (self.Mode.dir, self.Mode.ext):
            self.system.write_be_word(val, self.reg.ind_x)
            if mode == self.Mode.ext:
                self.system.tick()
        elif mode == self.Mode.idx:
            self.system.write_be_word(val+self.reg.ind_x, self.reg.ind_x)
            self.system.tick()
            self.system.tick()
        self.update_szv_word(self.reg.ind_x)
        self.system.tick()
        self.system.tick()

    def bsr(self, off):
        self.system.tick()
        self.return_stack.append((self.reg.ip, False))
        self.system.tick()
        self.system.write_be_word(self.reg.stack_ptr-1, self.reg.ip)
        self.system.tick()
        self.reg.stack_ptr = (self.reg.stack_ptr-2) & 0xFFFF
        self.system.tick()
        self.reg.ip += self.ip_offset(off)
        self.system.tick()
        self.system.tick()

    def jsr(self, mode, val):
        self.system.tick()
        self.return_stack.append((self.reg.ip, False))
        self.system.tick()
        self.system.write_be_word(self.reg.stack_ptr-1, self.reg.ip)
        self.system.tick()
        self.reg.stack_ptr = (self.reg.stack_ptr-2) & 0xFFFF
        self.system.tick()
        if mode == self.Mode.ext:
            self.reg.ip = val
        elif mode == self.Mode.idx:
            self.reg.ip = self.system.read_be_word(val+self.reg.ind_x)
            self.system.tick()
        self.system.tick()
        self.system.tick()
