from CPU import *


class System:
    def __init__(self, addr_bus_width=16):
        self.addr_bus_width = addr_bus_width
        self.addr_bus_mask = 2**addr_bus_width-1
        self.devices = {}
        self.ticks = 0
        self.read_watch = {}
        self.write_watch = {}

    def add(self, name, device):
        if isinstance(device, CPU):
            self.cpu = device
        else:
            self.devices[name] = device

    def reset(self):
        self.ticks = 0
        for device in self.devices.values():
            device.reset()
        self.cpu.reset()

    def run(self, endTick=0):
        checkBreak = False
        self.halt = False
        while not self.halt and (endTick == 0 or self.ticks < endTick):
            self.cpu.step(checkBreak)
            checkBreak = True

    def dump(self, addr, lines=4):
        per_line = 16

        def isprint(c):
            if c < 0x20 or c > 126:
                return False
            return True

        for line in range(0, lines):
            data = [self.read(addr+o) for o in range(0, per_line)]
            fmt = "{:04x}  ".format(addr)
            fmt += "{:02x} " * per_line
            pdata = ''
            for o in range(0, per_line):
                pdata += chr(data[o]) if isprint(data[o]) else '.'
            print(fmt.format(*data), pdata)
            addr += per_line

    def tick(self):
        self.ticks += 1
        for device in self.devices.values():
            device.tick(self.ticks)

    def read_be_word(self, addr):
        return self.read(addr) << 8 | self.read(addr+1)

    def write_be_word(self, addr, val):
        self.write(addr, val >> 8)
        self.write(addr+1, val & 0xFF)

    def read(self, addr):
        val = 0xFF
        bus_addr = addr & self.addr_bus_mask
        for device in self.devices.values():
            if device.addrInDevice(bus_addr):
                val = device.read(bus_addr)
                break
        if self.watched(self.read_watch, addr, val):
            print("** Read {:02x} from {:04x}".format(val, addr))
        return val

    def write(self, addr, val):
        if self.watched(self.write_watch, addr, val):
            print("** Write {:02x} to {:04x}".format(val, addr))
        addr &= self.addr_bus_mask
        for device in self.devices.values():
            if device.addrInDevice(addr):
                device.write(addr, val)
                return

    def addReadWatch(self, addr, match=None):
        self.read_watch[addr] = match

    def addWriteWatch(self, addr, match=None):
        self.write_watch[addr] = match

    def watched(self, watch_list, addr, val):
        if addr not in watch_list:
            return False

        match = watch_list[addr]
        if match is None:
            return True

        if 'val' in match and val != match['val']:
            return False

        if 'cpu_trace' in match:
            self.cpu.trace = match['cpu_trace']
        if 'halt' in match:
            self.halt = match['halt']
        return True
